//
// Mikael Klasson
// mklasson@acm.org
// http://mklasson.cjb.net
//
// Patches the Import section of an executable (containing a PE header) and 
// changes all references to a specific dll to refer to a dll of your choosing 
// instead -- e.g. iceptit.dll.
//
// 2003-10-04, Fluff, Initial version.

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include "../get_dll_functions/peheader.h"
using namespace std;

string tolower( string s ) {
	for( unsigned int j = 0; j < s.size(); ++j ) {
		if( isupper( s[j] ) ) {
			s[j] = tolower( s[j] );
		}
	}
	return s;
}

int main( int argc, char ** argv ) {
	FILE *		fin;
	string		exefile,dllname,iceptitname;
	unsigned int	upeoffset;
	unsigned int	udata;
	unsigned int	uimprva;
	unsigned int	uimpoffset;
	vector<string>	symbols;

	if( argc < 4 ) {
		cout << "Usage: patch_import <exefile> <dllname> <iceptit dll name>" << endl;
		cout << "Example: patch_import test.exe kernel32.dll iceptit.dll" << endl;
		return 1;
	}
	exefile = argv[1];
	dllname = argv[2];
	iceptitname = argv[3];

	cout << "Parsing " << exefile << endl;
	if( !( fin = fopen( exefile.c_str(), "rb+" ) ) ) {
		cout << "ERROR: opening " << exefile << endl;
		return 1;
	}

	fseek( fin, 0x3c, SEEK_SET );
	fread( &upeoffset, 4, 1, fin );
	fseek( fin, upeoffset, SEEK_SET );
	fread( &udata, 4, 1, fin );
	if( udata != 0x004550 ) {
		cout << "ERROR: no PE header" << endl;
		return 1;
	}

	fseek( fin, upeoffset + 0x80, SEEK_SET );
	fread( &uimprva, 4, 1, fin );

	//printf( "Import RVA: 0x%X\n", uimprva );
	
	if( !( uimpoffset = rva2phys( fin, uimprva ) ) ) {
		cout << "ERROR: couldn't locate import table" << endl;
		return 1;
	}
	//printf( "Import offset: 0x%X\n", uimpoffset );

	bool	found = false;
	bool	problem = false;
	for( int j = 0; ; ++j ) {
		// I obviously need to get newer info on the PE header.
		// My info says an Import Directory Entry is 24 bytes, but that file's from 1998...
		fseek( fin, uimpoffset + 20 * j + 12, SEEK_SET );
		fread( &udata, 4, 1, fin );
		if( !udata ) {
			break;
		}
		fseek( fin, rva2phys( fin, udata ), SEEK_SET );
		string	name;
		unsigned int	namewidth;
		for( namewidth = 0; ; ++namewidth ) {
			char c = fgetc( fin );
			if( !c ) {
				break;
			}
			name.push_back( c );
		}
		while( !fgetc( fin ) ) {
			++namewidth;
		}
		cout << "Imports " << name << ".";
		if( tolower( name ) == tolower( dllname ) ) {
			found = true;
			cout << " Patching." << endl;
			if( namewidth < iceptitname.size() ) {
				problem = true;
				cout << "  Error. I wanted to patch this, but \"" + iceptitname + "\" won't fit." << endl;
				cout << "  Rename " + iceptitname + " to something no more than " << namewidth << " characters long." << endl;
				cout << "  Just \"i\" is always a last resort." << endl;
			} else {
				fseek( fin, rva2phys( fin, udata ), SEEK_SET );
				fwrite( iceptitname.c_str(), iceptitname.size() + 1, 1, fin );
				for( unsigned int j = iceptitname.size() + 1; j < namewidth; ++j ) {
					fputc( 0, fin );
				}
				cout << "  Ok, now imports " << iceptitname << " instead." << endl;
			}
		} else {
			cout << " Ignoring." << endl;
		}
	}
	if( !found ) {
		cout << "Sorry, I couldn't find any reference to " << dllname << "." << endl;
	} else if( problem ) {
		cout << "There were errors while patching. Review the output above." << endl;
	} else {
		cout << "Everything went fine. Enjoy!" << endl;
	}
	fclose( fin );
}
